"""
Модуль для создания SSL контекста.
"""

import ssl
from typing import TypeAlias

from pydantic import BaseModel

StrOrBytesPath: TypeAlias = str | bytes  # stable


class CertificateSchema(BaseModel):
    """
    Схема необходимых файлов для создания SSL контекста.
    """

    ca_file: StrOrBytesPath
    cert_file: StrOrBytesPath
    key_file: StrOrBytesPath
    password: str | None = None


def make_ssl_context(params: CertificateSchema, check_hostname: bool = False) -> ssl.SSLContext:
    """
    Создаем SSL контекст.
    """
    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=params.ca_file)
    ssl_context.load_cert_chain(certfile=params.cert_file, keyfile=params.key_file, password=params.password)
    ssl_context.check_hostname = check_hostname
    return ssl_context



"""
Модуль, содержащий схемы файлового хранилища.
"""

from pathlib import Path

from pydantic import BaseModel


class S3StorageParamsSchema(BaseModel):
    """
    Параметры настроек для S3Storage.
    """

    """
    Параметры настроек для S3Storage.
    """

    endpoint: str
    aws_secret_access_key: str
    aws_access_key_id: str
    port: int
    bucket: str
    secure: bool = True
    region_name: str = 'us-east-1'


class LocalStorageParamsSchema(BaseModel):
    """
    Параметры настроек для LocalStorage.
    """

    path: Path


StorageParamsSchema = S3StorageParamsSchema | LocalStorageParamsSchema



"""
Модуль, содержащий модели.
"""

import datetime as dt

from sqlalchemy import DateTime
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.sql import func


class CreatedAtMixin:
    """
    Миксин, содержащий дату и время создания записи.
    """

    created_at: Mapped[dt.datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: dt.datetime.now(dt.UTC),
        server_default=func.now(),
    )


class UpdatedAtMixin:
    """
    Миксин, содержащий дату и время обновления записи.
    """

    updated_at: Mapped[dt.datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: dt.datetime.now(dt.UTC),
        server_default=func.now(),
        onupdate=lambda: dt.datetime.now(dt.UTC),
    )


class TimestampMixin(CreatedAtMixin, UpdatedAtMixin):
    """
    Миксин, содержащий дату и время создания и обновления записи.
    """



"""
Модуль, содержащий middleware.
"""

import time
from typing import Awaitable, Callable

from fastapi import FastAPI, Request, Response
from starlette.middleware.cors import CORSMiddleware


async def add_process_time_header(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
    start_time = time.perf_counter()
    response = await call_next(request)
    response.headers['x-process-time'] = f'{time.perf_counter() - start_time}'
    return response


def use_middleware(
    app: FastAPI,
    name: str,
    cors_origins: list[str],
    *,
    allow_methods: list[str] | None = None,
    allow_headers: list[str] | None = None,
) -> FastAPI:
    """
    Регистрируем middleware.
    """

    app.add_middleware(
        CORSMiddleware,
        allow_origins=cors_origins,
        allow_credentials=True,
        allow_methods=allow_methods or ['*'],
        allow_headers=allow_headers or ['*'],
    )

    app.middleware('http')(add_process_time_header)
    return app



"""
Модуль, содержащий вспомогательные функции для работы с Typer.
"""

import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import Any, ParamSpec, TypeVar

Param = ParamSpec('Param')
RetType = TypeVar('RetType')


def typer_async(func: Callable[Param, Coroutine[Any, Any, RetType]]) -> Callable[Param, RetType]:
    """
    Декоратор для асинхронного запуска Typer.
    """

    @wraps(func)
    def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
        return asyncio.run(func(*args, **kwargs))

    return wrapper



"""
Модуль, содержащий сервис распределенной блокировки.
"""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import AsyncContextManager, Protocol

from redis import asyncio as aioredis
from redis.exceptions import LockError as AIORedisLockError

from fast_clean.exceptions import LockError


class LockServiceProtocol(Protocol):
    """
    Протокол сервиса распределенной блокировки.
    """

    def lock(
        self,
        name: str,
        *,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
    ) -> AsyncContextManager[None]:
        """
        Осуществляем распределенную блокировку.
        """
        ...


class RedisLockService:
    """
    Сервис распределенной блокировки с помощью Redis.
    """

    def __init__(self, redis: aioredis.Redis) -> None:
        self.redis = redis

    @asynccontextmanager
    async def lock(
        self,
        name: str,
        *,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
    ) -> AsyncIterator[None]:
        """
        Осуществляем распределенную блокировку.
        """
        try:
            async with self.redis.lock(name, timeout=timeout, sleep=sleep, blocking_timeout=blocking_timeout):
                yield
        except AIORedisLockError as lock_error:
            raise LockError() from lock_error



"""
Пакет, содержащий схемы.
"""

from .exceptions import BusinessLogicExceptionSchema as BusinessLogicExceptionSchema
from .exceptions import ModelAlreadyExistsErrorSchema as ModelAlreadyExistsErrorSchema
from .exceptions import ValidationErrorSchema as ValidationErrorSchema
from .pagination import (
    AppliedPaginationResponseSchema as AppliedPaginationResponseSchema,
)
from .pagination import PaginationRequestSchema as PaginationRequestSchema
from .pagination import PaginationResponseSchema as PaginationResponseSchema
from .pagination import PaginationResultSchema as PaginationResultSchema
from .pagination import PaginationSchema as PaginationSchema
from .repository import CreateSchema as CreateSchema
from .repository import CreateSchemaInt as CreateSchemaInt
from .repository import ReadSchema as ReadSchema
from .repository import ReadSchemaInt as ReadSchemaInt
from .repository import UpdateSchema as UpdateSchema
from .repository import UpdateSchemaInt as UpdateSchemaInt
from .request_response import RemoteRequestSchema as RemoteRequestSchema
from .request_response import RemoteResponseSchema as RemoteResponseSchema
from .request_response import RequestSchema as RequestSchema
from .request_response import ResponseSchema as ResponseSchema



"""
Модуль, содержащий перечисления репозитория настроек.
"""

from enum import StrEnum, auto


class SettingsSourceEnum(StrEnum):
    """
    Источник настроек.
    """

    ENV = auto()
    """
    Настройки, получаемые из переменных окружения.
    """



"""
Пакет, содержащий репозиторий файлового хранилища.

Представлено 2 реализации:
- Local
- S3
"""

from collections.abc import AsyncIterator
from pathlib import Path
from typing import AsyncContextManager, Protocol, Self

from .enums import StorageTypeEnum
from .local import LocalStorageRepository
from .reader import AsyncStreamReaderProtocol as AsyncStreamReaderProtocol
from .reader import StreamReaderProtocol, StreamReadProtocol
from .s3 import S3StorageRepository
from .schemas import (
    LocalStorageParamsSchema,
    S3StorageParamsSchema,
    StorageParamsSchema,
)


class StorageRepositoryProtocol(Protocol):
    """
    Протокол репозитория файлового хранилища.
    """

    async def __aenter__(self: Self) -> Self:
        """
        Вход в контекст менеджера.
        """
        ...

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Выход из контекст менеджера.
        """
        ...

    async def exists(self: Self, path: str | Path) -> bool:
        """
        Проверяем существует ли файл.
        """
        ...

    async def listdir(self: Self, path: str | Path) -> list[str]:
        """
        Получаем список файлов и директорий в заданной директории.
        """
        ...

    async def is_file(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли файл по пути.
        """
        ...

    async def is_dir(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли директория по пути.
        """
        ...

    async def read(self: Self, path: str | Path) -> bytes:
        """
        Читаем содержимое файла.
        """
        ...

    def stream_read(self: Self, path: str | Path) -> AsyncContextManager[StreamReaderProtocol]:
        """
        Читаем содержимое файла в потоковом режиме.
        """
        ...

    async def write(self: Self, path: str | Path, content: str | bytes) -> None:
        """
        Создаем файл или переписываем существующий.
        """
        ...

    async def stream_write(
        self: Self,
        path: str | Path,
        stream: StreamReadProtocol,
    ) -> None:
        """
        Создаем файл или переписываем существующий в потоковом режиме.
        """
        ...

    def straming_read(self: Self, path: str | Path) -> AsyncIterator[bytes]:
        """
        Возвращаем асинхронные итератор потока байт.
        """
        ...

    async def delete(self: Self, path: str | Path) -> None:
        """
        Удаляем файл.
        """
        ...


class StorageRepositoryFactoryProtocol(Protocol):
    """
    Протокол фабрики репозиториев файлового хранилища.
    """

    async def make(self, storage_type: StorageTypeEnum, params: StorageParamsSchema) -> StorageRepositoryProtocol:
        """
        Создаем репозиторий файлового хранилища.
        """
        ...


class StorageRepositoryFactoryImpl:
    """
    Реализация фабрики репозиториев файлового хранилища.
    """

    async def make(self: Self, storage_type: StorageTypeEnum, params: StorageParamsSchema) -> StorageRepositoryProtocol:
        """
        Создаем репозиторий файлового хранилища.
        """
        if storage_type == StorageTypeEnum.S3 and isinstance(params, S3StorageParamsSchema):
            return S3StorageRepository(params)
        elif storage_type == StorageTypeEnum.LOCAL and isinstance(params, LocalStorageParamsSchema):
            return LocalStorageRepository(params)
        raise NotImplementedError()



"""
Пакет, содержащий репозиторий для выполнения CRUD операций над моделями.

Представлено две реализации:
- InMemory
- Db
"""

import uuid
from collections.abc import Iterable, Sequence
from typing import Protocol, Self

from fast_clean.schemas import PaginationResultSchema, PaginationSchema

from .db import DbCrudRepository as DbCrudRepository
from .db import DbCrudRepositoryInt as DbCrudRepositoryInt
from .in_memory import InMemoryCrudRepository as InMemoryCrudRepository
from .in_memory import InMemoryCrudRepositoryInt as InMemoryCrudRepositoryInt
from .type_vars import (
    CreateSchemaBaseType,
    CreateSchemaIntType,
    CreateSchemaType,
    IdTypeContravariant,
    ReadSchemaBaseType,
    ReadSchemaIntType,
    UpdateSchemaBaseType,
    UpdateSchemaIntType,
    UpdateSchemaType,
)


class CrudRepositoryBaseProtocol(
    Protocol[
        ReadSchemaBaseType,
        CreateSchemaBaseType,
        UpdateSchemaBaseType,
        IdTypeContravariant,
    ]
):
    """
    Протокол базового репозитория для выполнения CRUD операций над моделями.
    """

    async def get(self: Self, id: IdTypeContravariant) -> ReadSchemaBaseType:
        """
        Получаем модель по идентификатору.
        """
        ...

    async def get_or_none(self: Self, id: IdTypeContravariant) -> ReadSchemaBaseType | None:
        """
        Получаем модель или None по идентификатору.
        """
        ...

    async def get_by_ids(
        self: Self, ids: Sequence[IdTypeContravariant], *, exact: bool = False
    ) -> list[ReadSchemaBaseType]:
        """
        Получаем список моделей по идентификаторам.
        """
        ...

    async def get_all(self: Self) -> list[ReadSchemaBaseType]:
        """
        Получаем все модели.
        """
        ...

    async def paginate(
        self: Self,
        pagination: PaginationSchema,
        *,
        search: str | None = None,
        search_by: Iterable[str] | None = None,
        sorting: Iterable[str] | None = None,
    ) -> PaginationResultSchema[ReadSchemaBaseType]:
        """
        Получаем список моделей с пагинацией, поиском и сортировкой.
        """
        ...

    async def create(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем модель.
        """
        ...

    async def bulk_create(self: Self, create_objects: list[CreateSchemaBaseType]) -> list[ReadSchemaBaseType]:
        """
        Создаем несколько моделей.
        """
        ...

    async def update(self: Self, update_object: UpdateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Обновляем модель.
        """
        ...

    async def bulk_update(self: Self, update_objects: list[UpdateSchemaBaseType]) -> None:
        """
        Обновляем несколько моделей.
        """
        ...

    async def upsert(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем или обновляем модель.
        """
        ...

    async def delete(self: Self, ids: Sequence[IdTypeContravariant]) -> None:
        """
        Удаляем модели.
        """
        ...


class CrudRepositoryIntProtocol(
    CrudRepositoryBaseProtocol[
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
        int,
    ],
    Protocol[
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
    ],
):
    """
    Протокол репозитория для выполнения CRUD операций над моделями старого типа.
    """

    ...


class CrudRepositoryProtocol(
    CrudRepositoryBaseProtocol[ReadSchemaBaseType, CreateSchemaType, UpdateSchemaType, uuid.UUID],
    Protocol[ReadSchemaBaseType, CreateSchemaType, UpdateSchemaType],
):
    """
    Протокол репозитория для выполнения CRUD операций над моделями нового типа.
    """

    ...



"""
Модуль, содержащий схемы пагинации.
"""

from __future__ import annotations

from typing import Generic, Self, TypeVar

from pydantic import BaseModel, Field

from .request_response import ResponseSchema


class PaginationRequestSchema(BaseModel):
    """
    Схема запроса пагинации.
    """

    page: int = Field(gt=0)
    page_size: int = Field(gt=0)

    def to_pagination_schema(self: Self) -> PaginationSchema:
        """
        Преобразуем к схеме пагинации с помощью limit и offset.
        """
        return PaginationSchema(limit=self.page_size, offset=(self.page - 1) * self.page_size)


class AppliedPaginationResponseSchema(ResponseSchema):
    """
    Схема ответа примененной пагинации.
    """

    page: int
    page_size: int
    count: int


class PaginationResponseSchema(ResponseSchema):
    """
    Схема ответа пагинации.
    """

    pagination: AppliedPaginationResponseSchema


class PaginationSchema(BaseModel):
    """
    Схема пагинации с помощью limit и offset.
    """

    limit: int
    offset: int


T = TypeVar('T')


class PaginationResultSchema(BaseModel, Generic[T]):
    """
    Схема результата пагинации.
    """

    objects: list[T]
    count: int



"""
Пакет, содержащий репозиторий кеша.

Представлено две реализации:
- InMemory
- Redis
"""

from typing import ClassVar, Protocol, Self, cast

from fastapi_cache import FastAPICache
from redis import asyncio as aioredis

from fast_clean.settings import CoreCacheSettingsSchema

from .in_memory import InMemoryCacheRepository as InMemoryCacheRepository
from .redis import RedisCacheRepository as RedisCacheRepository


class CacheRepositoryProtocol(Protocol):
    """
    Протокол репозитория кеша.
    """

    async def get(self: Self, key: str) -> str | None:
        """
        Получаем значение.
        """
        ...

    async def set(self: Self, key: str, value: str, expire: int | None = None, nx: bool = False) -> None:
        """
        Устанавливаем значение.
        """
        ...

    async def get_with_ttl(self: Self, key: str) -> tuple[int, str | None]:
        """
        Получаем значение со сроком жизни.
        """
        ...

    async def incr(self: Self, key: str, amount: int = 1) -> int:
        """
        Инкрементируем значения.
        """
        ...

    async def decr(self: Self, key: str, amount: int = 1) -> int:
        """
        Декрементируем значения.
        """
        ...

    async def clear(self: Self, namespace: str | None = None, key: str | None = None) -> int:
        """
        Удаляем значение.
        """
        ...


class CacheManager:
    """
    Менеджер для работы с репозиторием кеша.
    """

    cache_repository: ClassVar[CacheRepositoryProtocol | None] = None

    @classmethod
    def init(cls, cache_settings: CoreCacheSettingsSchema):
        """
        Инициализируем кеш.
        """
        if cls.cache_repository is None:
            cache_backend: InMemoryCacheRepository | RedisCacheRepository
            match cache_settings.provider:
                case 'in_memory':
                    cache_backend = InMemoryCacheRepository()
                case 'redis':
                    if not cache_settings.redis:
                        raise ValueError('Redis not configured in settings')
                    cache_backend = RedisCacheRepository(
                        aioredis.from_url(url=str(cache_settings.redis.dsn), decode_responses=True)  # type: ignore
                    )
                case _:
                    raise ValueError('Cache is not initialized')
            FastAPICache.init(cache_backend, prefix=cache_settings.prefix)
            cls.cache_repository = cast(CacheRepositoryProtocol, cache_backend)
        return cls.cache_repository



"""
Модуль содержит имплементация работы с репозиторием по средствам протокола S3.
"""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Self, cast

import aiobotocore
import aiobotocore.session
from aiobotocore.response import StreamingBody
from aiobotocore.session import AioSession
from botocore.exceptions import ClientError

from fast_clean.repositories.storage.schemas import S3StorageParamsSchema

from .reader import AiofilesStreamReader, StreamReaderProtocol, StreamReadProtocol

if TYPE_CHECKING:
    from types_aiobotocore_s3.client import S3Client as AioBaseClient
else:
    from aiobotocore.client import AioBaseClient


class S3StorageRepository:
    """
    Репозиторий хранилища S3 с использованием aiobotocore.
    """

    def __init__(self, params: S3StorageParamsSchema) -> None:
        self.params = params
        self.bucket = self.params.bucket

        self.session: AioSession | None = None
        self.client: AioBaseClient | None = None

        protocol = 'https' if self.params.secure else 'http'
        self.endpoint_url = f'{protocol}://{self.params.endpoint}:{self.params.port}'

    async def __aenter__(self: Self) -> Self:
        self.session = aiobotocore.session.get_session()
        self.client = await self.session.create_client(
            's3',
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.params.aws_access_key_id,
            aws_secret_access_key=self.params.aws_secret_access_key,
            region_name=self.params.region_name,
        ).__aenter__()
        return self

    async def __aexit__(self: Self, exc_type, exc_val, exc_tb) -> None:
        """
        Выход из контектсного менеджера сессии.
        """
        if self.client:
            await self.client.__aexit__(exc_type, exc_val, exc_tb)
            self.client = None
            self.session = None

    async def exists(self: Self, path: str | Path) -> bool:
        assert self.client
        key = self.get_str_path(path)
        if key == '':
            key = '/'
        try:
            await self.client.head_object(Bucket=self.bucket, Key=key)
            return True
        except ClientError:
            return False

    async def listdir(self: Self, path: str | Path) -> list[str]:
        """
        Получаем список файлов и директорий в указанной лиректории.
        """
        assert self.client
        prefix = self.get_str_path(path)
        if prefix and not prefix.endswith('/'):
            prefix += '/'
        objects = []
        paginator = self.client.get_paginator('list_objects_v2')
        async for page in paginator.paginate(Bucket=self.bucket, Prefix=prefix, Delimiter='/'):
            if 'Contents' in page:
                objects.extend([obj.get('Key') for obj in page['Contents'] if obj.get('Key') != prefix])
            if 'CommonPrefixes' in page:
                objects.extend([folder.get('Prefix') for folder in page['CommonPrefixes']])
        return objects

    async def is_file(self: Self, path: str | Path) -> bool:
        """
        Проверяем, является ли путь файлом.
        """
        return await self.exists(path)

    async def is_dir(self: Self, path: str | Path) -> bool:
        """
        Проверяем, является ли путь дирректорией.
        """
        assert self.client
        prefix = self.get_str_path(path)
        if prefix != '' and not prefix.endswith('/'):
            prefix += '/'
        response = await self.client.list_objects_v2(
            Bucket=self.bucket,
            Prefix=prefix,
            MaxKeys=1,
            Delimiter='/',
        )
        return 'Contents' in response or 'CommonPrefixes' in response

    async def read(self: Self, path: str | Path) -> bytes:
        """
        Читаем содержимое файла.
        """
        assert self.client
        key = self.get_str_path(path)
        response = await self.client.get_object(Bucket=self.bucket, Key=key)
        async with response['Body'] as stream:
            return await stream.read()

    async def write(self: Self, path: str | Path, content: str | bytes) -> None:
        """
        Создаем файл или перезаписываем существующий.
        """
        assert self.client
        key = self.get_str_path(path)
        content = content.encode('utf-8') if isinstance(content, str) else content
        await self.client.put_object(Bucket=self.bucket, Key=key, Body=content)

    @asynccontextmanager
    async def stream_read(self: Self, path: str | Path) -> AsyncIterator[StreamReaderProtocol]:
        """
        Читаем содержимое в потоковом режиме.
        """
        assert self.client
        key = self.get_str_path(path)
        response = await self.client.get_object(Bucket=self.bucket, Key=key)
        yield AiofilesStreamReader(response['Body'])

    async def straming_read(self: Self, path: str | Path) -> AsyncIterator[bytes]:
        assert self.client
        key = self.get_str_path(path)
        response = await self.client.get_object(Bucket=self.bucket, Key=key)
        async for chunk in response['Body']:
            yield chunk

    async def stream_write(self: Self, path: str | Path, stream: StreamReadProtocol) -> None:
        """
        Создаем поток на запись файла и перезаписываем существующий, если он есть.
        """
        assert self.client
        await self.client.put_object(Bucket=self.bucket, Key=self.get_str_path(path), Body=cast(StreamingBody, stream))

    async def delete(self: Self, path: str | Path) -> None:
        """
        Удаление файла.
        """
        assert self.client
        key = self.get_str_path(path)
        await self.client.delete_object(Bucket=self.bucket, Key=key)

    @staticmethod
    def get_str_path(path: str | Path) -> str:
        """
        Получаем путь в виде строки.
        """
        if isinstance(path, Path):
            return '' if path == Path('') else str(path)
        return path



"""
Модуль, содержащий схемы запросов и ответов.
"""

from pydantic import AliasGenerator, BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class RequestSchema(BaseModel):
    """
    Схема запроса.
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            validation_alias=to_camel,
        )
    )


RemoteResponseSchema = RequestSchema


class ResponseSchema(BaseModel):
    """
    Схема ответа.
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            serialization_alias=to_camel,
        )
    )


RemoteRequestSchema = ResponseSchema



"""
Модуль для работы со строками.
"""

import base64
import string
from random import choice


def make_random_string(size: int) -> str:
    """
    Создаем случайную строку.
    """
    return ''.join(choice(string.ascii_letters + string.digits) for _ in range(size))


def encode_base64(raw_value: str) -> str:
    """
    Кодируем строку в base64.
    """
    return base64.b64encode(raw_value.encode()).decode()


def decode_base64(value: str) -> str:
    """
    Декодируем строку из base64.
    """
    return base64.b64decode(value).decode()



"""
Модуль, содержащий функции для преобразования типов.
"""

true_set = {'yes', 'true', 't', 'y', '1'}
false_set = {'no', 'false', 'f', 'n', '0'}


def str_to_bool(value: str) -> bool:
    """
    Преобразуем строку к логическом значению.
    """
    value = value.lower()
    if value in true_set:
        return True
    if value in false_set:
        return False
    raise ValueError('Expected "%s"' % '", "'.join(true_set | false_set))



"""
Модуль, содержащий тесты функционала, связанного с брокером сообщений.
"""

import ssl

from fast_clean.broker import BrokerCredentials
from fast_clean.settings import CoreKafkaSettingsSchema
from faststream.security import BaseSecurity, SASLPlaintext
from pytest_mock import MockFixture

from tests.settings import SettingsSchema


class TestBrokerCredentials:
    """
    Тесты класса получения параметров для подключения к Kafka.
    """

    @staticmethod
    def test_none_credentials(settings: SettingsSchema) -> None:
        """
        Тестируем получение отсутствующих параметров для подключения.
        """
        credentials = BrokerCredentials(
            CoreKafkaSettingsSchema(
                bootstrap_servers=settings.kafka.bootstrap_servers, group_id=settings.kafka.group_id
            )
        )
        assert credentials.get() is None

    @staticmethod
    def test_ssl_credentials(settings: SettingsSchema, mocker: MockFixture) -> None:
        """
        Тестируем получение SSL параметров для подключения.
        """
        kafka_settings = CoreKafkaSettingsSchema(
            group_id=settings.kafka.group_id,
            bootstrap_servers=settings.kafka.bootstrap_servers,
            credentials='SSL',
            ca_file='path/to/ca.pem',
            cert_file='path/to/cert.pem',
            key_file='path/to/key.pem',
            password='test_password',
        )
        mock_create_default_context = mocker.patch('ssl.create_default_context')
        credentials = BrokerCredentials(kafka_settings)
        security = credentials.get()
        mock_create_default_context.assert_called_once_with(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=kafka_settings.ca_file
        )
        mock_ssl_context = mock_create_default_context.return_value
        mock_ssl_context.load_cert_chain.assert_called_once_with(
            certfile=kafka_settings.cert_file, keyfile=kafka_settings.key_file, password=kafka_settings.password
        )
        assert isinstance(security, BaseSecurity)
        assert security.use_ssl
        ssl_context = security.ssl_context
        assert ssl_context is not None
        assert ssl_context == mock_ssl_context
        assert not ssl_context.check_hostname

    @staticmethod
    def test_sasl_credentials(settings: SettingsSchema) -> None:
        """
        Тестируем получение SASL параметров для подключения.
        """
        kafka_settings = CoreKafkaSettingsSchema(
            group_id=settings.kafka.group_id,
            bootstrap_servers=settings.kafka.bootstrap_servers,
            credentials='SASL',
            broker_username='test_user',
            broker_password='test_password',
        )
        credentials = BrokerCredentials(kafka_settings)
        security = credentials.get()
        assert isinstance(security, SASLPlaintext)
        assert security.use_ssl
        assert security.username == kafka_settings.broker_username
        assert security.password == kafka_settings.broker_password



import logging

import sentry_sdk
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.logging import LoggingIntegration

from .enums import EnvironmentEnum


def use_sentry(
    dsn: str | None,
    environment: EnvironmentEnum = EnvironmentEnum.DEVELOPMENT,
    level: int = logging.DEBUG,
    event_level: int = logging.ERROR,
) -> None:
    sentry_logging = LoggingIntegration(
        level=level,
        event_level=event_level,
    )
    sentry_sdk.init(dsn=dsn, environment=str(environment), integrations=[sentry_logging, FastApiIntegration()])



"""
Модуль, содержащий реализацию сервиса криптографии с использованием алгоритма AES.
"""

import base64
import os
import warnings
from typing import Self

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class AesGcmCryptographyService:
    """
    Сервис криптографии с использованием алгоритма AES в режиме GCM.
    """

    def __init__(self, secret_key: str) -> None:
        self.secret_key = secret_key
        self.backend = default_backend()

    def encrypt(self: Self, data: str) -> str:
        """
        Зашифровываем данные.
        """
        bytes_data = data.encode('utf-8')
        iv = os.urandom(12)
        cipher = Cipher(algorithms.AES(self.key), modes.GCM(iv), backend=self.backend)
        encryptor = cipher.encryptor()
        cipher_text = encryptor.update(bytes_data) + encryptor.finalize()
        assert hasattr(encryptor, 'tag')
        encrypted_data = iv + cipher_text + encryptor.tag
        return base64.b64encode(encrypted_data).decode('ascii')

    def decrypt(self: Self, encrypted_data: str) -> str:
        """
        Расшифровываем данные.
        """
        bytes_encrypted_data = base64.b64decode(encrypted_data)
        iv = bytes_encrypted_data[:12]
        tag = bytes_encrypted_data[-16:]
        cipher_text = bytes_encrypted_data[12:-16]
        cipher = Cipher(algorithms.AES(self.key), modes.GCM(iv, tag), backend=self.backend)
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(cipher_text) + decryptor.finalize()
        return decrypted_data.decode('utf-8')

    @property
    def key(self: Self) -> bytes:
        """
        Ключ длиной в 32 байта.
        """
        key_bytes = self.secret_key.encode()
        if len(key_bytes) > 32:
            return key_bytes[:32]
        if len(key_bytes) < 32:
            diff = 32 - len(key_bytes)
            return key_bytes + b'0' * diff
        return key_bytes


class AesCbcCryptographyService:
    """
    Сервис криптографии с использованием алгоритма AES в режиме CBC.

    Данный класс использует режим CBC без аутентификации и поэтому может быть небезопасным.
    В будущем данный класс будет удален.
    https://sonarsource.github.io/rspec/#/rspec/S5542/python
    """

    def __init__(self, secret_key: str) -> None:
        warnings.warn(
            f'{self.__class__.__name__} is deprecated and will be removed in future versions',
            DeprecationWarning,
            stacklevel=2,
        )
        self.secret_key = secret_key
        self.backend = default_backend()

    def encrypt(self: Self, data: str) -> str:
        """
        Зашифровываем данные.
        """
        bytes_data = data.encode(encoding='utf-8')
        iv = os.urandom(16)
        cipher = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=self.backend)
        encryptor = cipher.encryptor()
        padder = padding.PKCS7(128).padder()
        padded_data = padder.update(bytes_data) + padder.finalize()
        cipher_text = encryptor.update(padded_data) + encryptor.finalize()
        return base64.b64encode(iv + cipher_text).decode('ascii')

    def decrypt(self: Self, encrypted_data: str) -> str:
        """
        Расшифровываем данные.
        """
        bytes_encrypted_data = base64.b64decode(encrypted_data)
        iv = bytes_encrypted_data[:16]
        cipher_text = bytes_encrypted_data[16:]
        cipher = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=self.backend)
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(cipher_text) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        unpadded_data = unpadder.update(decrypted_data) + unpadder.finalize()
        return unpadded_data.decode(encoding='utf-8')

    @property
    def key(self: Self) -> bytes:
        """
        Ключ длиной в 32 байта.
        """
        key_bytes = self.secret_key.encode()
        if len(key_bytes) > 32:
            return key_bytes[:32]
        if len(key_bytes) < 32:
            diff = 32 - len(key_bytes)
            return key_bytes + b'0' * diff
        return key_bytes



"""
Пакет, содержащий вспомогательные функции и классы.
"""

from .process import run_in_processpool as run_in_processpool
from .pydantic import rebuild_schemas as rebuild_schemas
from .ssl_context import CertificateSchema as CertificateSchema
from .ssl_context import make_ssl_context as make_ssl_context
from .string import decode_base64 as decode_base64
from .string import encode_base64 as encode_base64
from .string import make_random_string as make_random_string
from .thread import run_in_threadpool as run_in_threadpool
from .time import ts_now as ts_now
from .type_converters import str_to_bool as str_to_bool
from .typer import typer_async as typer_async



"""
Модуль, содержащий функционал, связанный с базой данных.
"""

from __future__ import annotations

import uuid
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, AsyncContextManager, Protocol, Self

import sqlalchemy as sa
from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import (
    AsyncAttrs,
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.sql import func
from sqlalchemy_utils.types import UUIDType

from .settings import CoreDbSettingsSchema, CoreSettingsSchema

if TYPE_CHECKING:
    from .repositories import SettingsRepositoryProtocol

POSTGRES_INDEXES_NAMING_CONVENTION = {
    'ix': '%(column_0_label)s_idx',
    'uq': '%(table_name)s_%(column_0_name)s_key',
    'ck': '%(table_name)s_%(constraint_name)s_check',
    'fk': '%(table_name)s_%(column_0_name)s_fkey',
    'pk': '%(table_name)s_pkey',
}

metadata = MetaData(naming_convention=POSTGRES_INDEXES_NAMING_CONVENTION)


def make_async_engine(
    db_dsn: str,
    *,
    scheme: str = 'public',
    echo: bool = False,
    pool_pre_ping: bool = True,
    disable_prepared_statements: bool = True,
) -> AsyncEngine:
    """
    Создаем асинхронный движок.
    """
    connect_args: dict[str, Any] = {}
    if disable_prepared_statements:
        connect_args['prepare_threshold'] = None
    return create_async_engine(
        db_dsn,
        echo=echo,
        pool_pre_ping=pool_pre_ping,
        connect_args=connect_args,
    )


def make_async_session_factory(
    db_dsn: str,
    *,
    scheme: str = 'public',
    echo: bool = False,
    pool_pre_ping: bool = True,
    disable_prepared_statements: bool = True,
) -> async_sessionmaker[AsyncSession]:
    """
    Создаем фабрику асинхронных сессий.
    """
    asyncio_engine = make_async_engine(
        db_dsn,
        scheme=scheme,
        echo=echo,
        pool_pre_ping=pool_pre_ping,
        disable_prepared_statements=disable_prepared_statements,
    )
    return async_sessionmaker(asyncio_engine, expire_on_commit=False, autoflush=False)


class Base(AsyncAttrs, DeclarativeBase):
    """
    Базовая родительская модель.
    """

    __abstract__ = True

    metadata = metadata


class BaseUUID(Base):
    """
    Базовая родительская модель нового типа.
    """

    __abstract__ = True

    id: Mapped[uuid.UUID] = mapped_column(
        UUIDType(binary=False),
        primary_key=True,
        default=uuid.uuid4,
        server_default=func.gen_random_uuid(),
    )


class BaseInt(Base):
    """
    Базовая родительская модель старого типа.
    """

    __abstract__ = True

    id: Mapped[int] = mapped_column(primary_key=True)


class SessionFactory:
    """
    Фабрика сессий.
    """

    async_session_factory: async_sessionmaker[AsyncSession] | None = None

    @classmethod
    @asynccontextmanager
    async def make_async_session_static(
        cls, settings_repository: SettingsRepositoryProtocol
    ) -> AsyncIterator[AsyncSession]:
        """
        Создаем асинхронную сессию с помощью статической фабрики.
        """
        if cls.async_session_factory is None:
            cls.async_session_factory = await cls.make_async_session_factory(settings_repository)
        async with cls.async_session_factory() as session:
            yield session

    @classmethod
    async def make_async_session_dynamic(
        cls, settings_repository: SettingsRepositoryProtocol
    ) -> async_sessionmaker[AsyncSession]:
        """
        Создаем асинхронную сессию с помощью динамической фабрики.
        """
        return await cls.make_async_session_factory(settings_repository)

    @staticmethod
    async def make_async_session_factory(
        settings_repository: SettingsRepositoryProtocol,
    ) -> async_sessionmaker[AsyncSession]:
        """
        Создаем фабрику асинхронных сессий.
        """
        settings = await settings_repository.get(CoreSettingsSchema)
        db_settings = await settings_repository.get(CoreDbSettingsSchema)
        return make_async_session_factory(
            db_settings.dsn,
            scheme=db_settings.scheme,
            echo=settings.debug,
            pool_pre_ping=db_settings.pool_pre_ping,
            disable_prepared_statements=db_settings.disable_prepared_statements,
        )


class SessionManagerProtocol(Protocol):
    """
    Протокол менеджера сессий.
    """

    def get_session(self: Self, immediate: bool = True) -> AsyncContextManager[AsyncSession]:
        """
        Получаем сессию для выполнения запроса.
        """
        ...


class SessionManagerImpl:
    """
    Реализация менеджера сессий.
    """

    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    @asynccontextmanager
    async def get_session(self: Self, immediate: bool = True) -> AsyncIterator[AsyncSession]:
        """
        Получаем сессию для выполнения запроса.
        """
        if self.session.in_transaction():
            yield self.session
        else:
            async with self.session.begin():
                if immediate:
                    await self.session.execute(sa.text('SET CONSTRAINTS ALL IMMEDIATE'))
                yield self.session



"""
Модуль, содержащий зависимости.
"""

import json
from collections.abc import AsyncIterator, Sequence
from typing import Annotated

from dishka import Provider, Scope, provide
from fastapi import Depends, Request
from faststream.kafka import KafkaBroker
from flatten_dict import unflatten
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker
from starlette.datastructures import FormData
from stringcase import snakecase

from .broker import BrokerFactory
from .db import SessionFactory, SessionManagerImpl, SessionManagerProtocol
from .repositories import (
    CacheManager,
    CacheRepositoryProtocol,
    LocalStorageParamsSchema,
    S3StorageParamsSchema,
    SettingsRepositoryFactoryImpl,
    SettingsRepositoryFactoryProtocol,
    SettingsRepositoryProtocol,
    SettingsSourceEnum,
    StorageRepositoryFactoryImpl,
    StorageRepositoryFactoryProtocol,
    StorageRepositoryProtocol,
    StorageTypeEnum,
)
from .schemas import PaginationRequestSchema
from .services import (
    CryptographicAlgorithmEnum,
    CryptographyServiceFactory,
    CryptographyServiceProtocol,
    LockServiceProtocol,
    RedisLockService,
    SeedService,
    TransactionService,
)
from .settings import CoreCacheSettingsSchema, CoreKafkaSettingsSchema, CoreSettingsSchema, CoreStorageSettingsSchema


async def get_nested_form_data(request: Request) -> FormData:
    """
    Получаем форму, позволяющую использовать вложенные словари.
    """
    dot_data = {k.replace('[', '.').replace(']', ''): v for k, v in (await request.form()).items()}
    nested_data = unflatten(dot_data, 'dot')
    for k, v in nested_data.items():
        if isinstance(v, dict):
            nested_data[k] = json.dumps(v)
    return FormData(nested_data)


def get_pagination(page: int | None = None, page_size: int | None = None) -> PaginationRequestSchema:
    """
    Получаем входные данные пагинации.
    """
    return PaginationRequestSchema(page=page or 1, page_size=page_size or 10)


def get_sorting(sorting: str | None = None) -> Sequence[str]:
    """
    Получаем входные данные сортировки.
    """
    if not sorting:
        return []
    return [s[0] + snakecase(s[1:]) if s[0] == '-' else snakecase(s) for s in sorting.split(',')]


NestedFormData = Annotated[FormData, Depends(get_nested_form_data)]
Pagination = Annotated[PaginationRequestSchema, Depends(get_pagination)]
Sorting = Annotated[Sequence[str], Depends(get_sorting)]


class CoreProvider(Provider):
    """
    Провайдер зависимостей.
    """

    scope = Scope.REQUEST

    # --- repositories ---

    settings_repository_factory = provide(
        SettingsRepositoryFactoryImpl, provides=SettingsRepositoryFactoryProtocol, scope=Scope.APP
    )
    storage_repository_factory = provide(
        StorageRepositoryFactoryImpl, provides=StorageRepositoryFactoryProtocol, scope=Scope.APP
    )

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_settings_repository(
        settings_repository_factory: SettingsRepositoryFactoryProtocol,
    ) -> SettingsRepositoryProtocol:
        """
        Получаем репозиторий настроек.
        """
        return await settings_repository_factory.make(SettingsSourceEnum.ENV)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_settings(settings_repository: SettingsRepositoryProtocol) -> CoreSettingsSchema:
        """
        Получаем настройки.
        """
        return await settings_repository.get(CoreSettingsSchema)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_cache_settings(settings_repository: SettingsRepositoryProtocol) -> CoreCacheSettingsSchema:
        """
        Получаем настройки кеша.
        """
        return await settings_repository.get(CoreCacheSettingsSchema)

    @provide
    @staticmethod
    async def get_broker_repository(settings_repository: SettingsRepositoryProtocol) -> AsyncIterator[KafkaBroker]:
        """
        Получаем репозиторий брокера сообщений.
        """
        kafka_settings = await settings_repository.get(CoreKafkaSettingsSchema)
        yield BrokerFactory.make_static(kafka_settings)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_cache_repository(settings_repository: SettingsRepositoryProtocol) -> CacheRepositoryProtocol:
        """
        Получаем репозиторий кеша.
        """
        cache_settings = await settings_repository.get(CoreCacheSettingsSchema)
        return CacheManager.init(cache_settings)

    @provide
    @staticmethod
    async def get_storage_repository(
        settings_repository: SettingsRepositoryProtocol,
        storage_repository_factory: StorageRepositoryFactoryProtocol,
    ) -> AsyncIterator[StorageRepositoryProtocol]:
        """
        Получаем репозиторий файлового хранилища.
        """
        storage_settings = await settings_repository.get(CoreStorageSettingsSchema)
        if storage_settings.provider == 's3' and storage_settings.s3 is not None:
            storage_repository = await storage_repository_factory.make(
                StorageTypeEnum.S3,
                S3StorageParamsSchema.model_validate(storage_settings.s3.model_dump()),
            )
            async with storage_repository:
                yield storage_repository
        elif storage_settings.provider == 'local':
            storage_repository = await storage_repository_factory.make(
                StorageTypeEnum.LOCAL, LocalStorageParamsSchema(path=storage_settings.dir)
            )
            async with storage_repository:
                yield storage_repository
        else:
            raise NotImplementedError(f'Storage {storage_settings.provider} is not allowed')

    # --- db ---

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_async_sessionmaker(
        settings_repository: SettingsRepositoryProtocol,
    ) -> async_sessionmaker[AsyncSession]:
        return await SessionFactory.make_async_session_dynamic(settings_repository)

    @provide
    @staticmethod
    async def get_async_session(session_maker: async_sessionmaker[AsyncSession]) -> AsyncIterator[AsyncSession]:
        """
        Получаем асинхронную сессию.
        """
        async with session_maker() as session:
            yield session

    @provide
    @staticmethod
    def get_session_manager(session: AsyncSession) -> SessionManagerProtocol:
        """
        Получаем менеджер сессий.
        """
        return SessionManagerImpl(session)

    # --- services ---

    seed_service = provide(SeedService, scope=Scope.REQUEST)
    transaction_service = provide(TransactionService)

    @provide(scope=Scope.APP)
    @staticmethod
    def get_cryptography_service_factory(settings: CoreSettingsSchema) -> CryptographyServiceFactory:
        """
        Получаем фабрику сервисов криптографии.
        """
        return CryptographyServiceFactory(settings.secret_key)

    @provide
    @staticmethod
    async def get_cryptography_service(
        cryptography_service_factory: CryptographyServiceFactory,
    ) -> CryptographyServiceProtocol:
        """
        Получаем сервис криптографии.
        """
        return await cryptography_service_factory.make(CryptographicAlgorithmEnum.AES_GCM)

    @provide(scope=Scope.APP)
    @staticmethod
    def get_lock_service(cache_settings: CoreCacheSettingsSchema) -> LockServiceProtocol:
        """
        Получаем сервис распределенной блокировки.
        """
        redis_client = CacheManager.init(cache_settings)
        return RedisLockService(redis_client)  # type: ignore


provider = CoreProvider()



"""
Модуль, содержащий функционал, связанный логированием.
"""

import logging
import logging.config
from pathlib import Path

import yaml


def use_logging(base_dir: Path) -> None:
    """
    Применяем настройки логирования.
    """
    files = ['.logging.dev.yaml', '.logging.yaml']
    for file in files:
        ls_file = base_dir / file
        if ls_file.exists():
            with open(ls_file, 'rt') as f:
                config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)
            break
    else:
        print(f'Missing configuration logging files: {", ".join(files)}')
        exit(0)



"""
Модуль, содержащий исключения репозитория настроек.
"""


class SettingsRepositoryError(Exception):
    """
    Ошибка репозитория настроек.
    """

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message



"""
Модуль, содержащий сервис для загрузки данных из файлов.
"""

import importlib
import json
import os
from pathlib import Path
from typing import Any, cast

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.ext.asyncio import AsyncSession

from ..db import SessionManagerProtocol


class SeedService:
    """
    Реализация сервиса для загрузки данных из файлов.
    """

    def __init__(self, session_manager: SessionManagerProtocol) -> None:
        self.session_manager = session_manager

    async def load_data(self, directory: str | Path | None = None) -> None:
        """
        Загружаем данные из файлов по пути.
        """
        directory = directory if directory is not None else self.find_directory()
        directory = Path(directory) if isinstance(directory, str) else directory
        async with self.session_manager.get_session() as session:
            for file in sorted(os.listdir(directory)):
                file_path = directory / file
                with open(file_path) as f:
                    items = json.load(f)
                    for item in items:
                        await self.upsert_item(item, session)

    @staticmethod
    def find_directory() -> Path:
        """
        Ищем директорию с файлами для загрузки.
        """
        cwd = Path(os.getcwd())
        virtual_env_paths = {path.parent for path in cwd.rglob('pyvenv.cfg')}
        for path in cwd.rglob('seed'):
            if not any(path.is_relative_to(venv) for venv in virtual_env_paths):
                return path
        raise ValueError('Seed directory not found')

    @classmethod
    async def upsert_item(cls, item: Any, session: AsyncSession) -> None:
        """
        Сохраняем запись в базу данных.
        """
        model_type = cls.import_from_string(item['model'])
        primary_keys = {key.name for key in cast(Any, sa.inspect(model_type)).primary_key}
        values = {**item['fields']}
        item_id = item.get('id')
        if item_id:
            values['id'] = item_id
        await session.execute(
            insert(model_type)
            .values(values)
            .on_conflict_do_update(
                index_elements=primary_keys,
                set_={k: v for k, v in values.items() if k not in primary_keys},
            )
        )

    @staticmethod
    def import_from_string(import_str: str) -> sa.TableClause:
        """
        Импортируем таблицу.
        """
        package_name, model_name = import_str.rsplit('.', maxsplit=1)
        package = importlib.import_module(package_name)
        return getattr(package, model_name)



import tomllib
from functools import lru_cache
from pathlib import Path
from typing import Any

from pydantic import BaseModel

__all__ = (
    'use_toml_info',
    'ProjectInfo',
)


class ProjectInfo(BaseModel):
    """
    Схема для получения информации о проекте.
    """

    name: str
    version: str
    description: str | None = None


def use_toml(dir: Path) -> dict[str, Any]:
    with open(Path(dir) / 'pyproject.toml', 'rb') as f:
        return tomllib.load(f)


@lru_cache(maxsize=1)
def use_toml_info(dir: Path) -> ProjectInfo:
    """
    Получение версии приложения из pyproject.toml.
    """
    return ProjectInfo.model_validate(use_toml(dir)['project'])



"""
Модуль запуска тяжелых операций в ProcessPoolExecutor.
"""

import asyncio
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from typing import Callable, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec('P')
R = TypeVar('R')


process_pool: ProcessPoolExecutor | None = None


async def run_in_processpool(fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    """
    Запуск функции в отдельном процессе.

    Используем fork в связи с https://github.com/python/cpython/issues/94765.
    """
    global process_pool
    if process_pool is None:
        process_pool = ProcessPoolExecutor(mp_context=mp.get_context('fork'))
    kwargs_fn = partial(fn, *args, **kwargs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(process_pool, kwargs_fn)



"""
Пакет, содержащий сервисы.
"""

from .cryptography import AesGcmCryptographyService as AesGcmCryptographyService
from .cryptography import CryptographicAlgorithmEnum as CryptographicAlgorithmEnum
from .cryptography import CryptographyServiceFactory as CryptographyServiceFactory
from .cryptography import CryptographyServiceProtocol as CryptographyServiceProtocol
from .lock import LockServiceProtocol as LockServiceProtocol
from .lock import RedisLockService as RedisLockService
from .seed import SeedService as SeedService
from .transaction import TransactionService as TransactionService
